import { InjectionToken, Injectable, ɵɵdefineInjectable, ɵɵinject, Inject, Directive, TemplateRef, Optional, ViewContainerRef, SkipSelf, Host, Input, NgModule } from '@angular/core';
import { Observable, Subject, of, from } from 'rxjs';
import { take, switchMap, takeUntil, map, tap } from 'rxjs/operators';
import { Router } from '@angular/router';

function flatten(config, { parse = (v) => v, group = false } = {}) {
    const flatConfig = {};
    function setConfig(path, value) {
        if (!flatConfig[path]) {
            flatConfig[path] = new Set();
        }
        flatConfig[path].add(value);
    }
    function getPath(path, delimiter, prop) {
        return path ? path + delimiter + prop : prop;
    }
    function visitor(accesses, prop, value, path) {
        if (Array.isArray(value)) {
            value.forEach((access) => {
                visitor(accesses, prop, access, path);
            });
        }
        else if (typeof value === 'object') {
            const childrenAccesses = children(value, getPath(path, '.', prop));
            if (group) {
                accesses = accesses.concat(childrenAccesses);
                accesses.forEach((access) => {
                    setConfig(getPath(path, ':', access.prop), access.action);
                });
            }
        }
        else {
            const expression = parse(value.replace(/\s/g, ''));
            setConfig(getPath(path, ':', prop), expression);
            accesses = accesses.concat({ action: expression, prop });
        }
        return accesses;
    }
    function children(obj, path = '') {
        return Object.keys(obj).reduce((accesses, prop) => visitor(accesses, prop, obj[prop], path), []);
    }
    children(config);
    return flatConfig;
}

const TokenType = {
    PAR_OPEN: '('.charCodeAt(0),
    PAR_CLOSE: ')'.charCodeAt(0),
    OP_NOT: '!'.charCodeAt(0),
    BINARY_AND: '&'.charCodeAt(0),
    BINARY_OR: '|'.charCodeAt(0),
    LITERAL: 'LITERAL',
    END: 'END',
    LEAF: 'LEAF',
    ATOMIC: 'ATOMIC',
};

const operator = (evaluator$, op) => (source) => {
    return new Observable((observer) => {
        const finish$ = new Subject();
        let lastEval;
        finish$.pipe(take(1)).subscribe((value) => {
            observer.next(value);
            observer.complete();
        }, (e) => observer.error(e));
        source.pipe(switchMap(evaluator$), takeUntil(finish$)).subscribe((value) => {
            lastEval = op === TokenType.OP_NOT ? !value : value;
            if (op === TokenType.BINARY_AND && !lastEval) {
                finish$.next(false);
            }
            else if (op === TokenType.BINARY_OR && lastEval) {
                finish$.next(true);
            }
        }, (e) => {
            finish$.error(e);
        }, () => {
            finish$.next(lastEval);
        });
    });
};

const Tokenizer = (exp) => {
    let literal = '';
    const tokens = [];
    for (const char of exp) {
        const code = char.charCodeAt(0);
        switch (code) {
            case TokenType.PAR_OPEN:
            case TokenType.PAR_CLOSE:
            case TokenType.OP_NOT:
            case TokenType.BINARY_AND:
            case TokenType.BINARY_OR:
                if (literal) {
                    tokens.push({
                        type: TokenType.LITERAL,
                        value: literal,
                    });
                    literal = '';
                }
                tokens.push({
                    type: code,
                    value: char,
                });
                break;
            default:
                literal += char;
        }
    }
    if (literal)
        tokens.push({
            type: TokenType.LITERAL,
            value: literal,
        });
    return tokens;
};
const ɵ0 = Tokenizer;

const PolishNotation = (tokens) => {
    const stack = [];
    let queue = [];
    let subQueue = [];
    let subStack = [];
    tokens.forEach((token) => {
        switch (token.type) {
            case TokenType.LITERAL:
            case TokenType.OP_NOT:
                if (subStack.length) {
                    subQueue.push(token);
                }
                else {
                    queue.push(token);
                }
                break;
            case TokenType.BINARY_AND:
            case TokenType.BINARY_OR:
                if (subStack.length) {
                    subStack.unshift(token);
                }
                else {
                    stack.unshift(token);
                }
                break;
            case TokenType.PAR_OPEN:
                subStack.unshift(token);
                break;
            case TokenType.PAR_CLOSE:
                while (subStack.length &&
                    subStack[subStack.length - 1].type !== TokenType.PAR_OPEN) {
                    subQueue.unshift(subStack.pop());
                }
                subStack.pop();
                if (subStack.length &&
                    subStack[subStack.length - 1].type === TokenType.OP_NOT) {
                    subQueue.unshift(subStack.pop());
                }
                queue = queue.concat([...subStack, ...subQueue]);
                subStack = [];
                subQueue = [];
                break;
            default:
                break;
        }
    });
    const result = (stack.length && [...stack, ...queue]) || queue;
    return result;
};
const ɵ0$1 = PolishNotation;
const PolishGenerator = function* (polish) {
    for (let index = 0; index < polish.length - 1; index++) {
        yield polish[index];
    }
    return polish[polish.length - 1];
};
const ɵ1 = PolishGenerator;

class ExpNode {
    constructor(op, left, right, literal) {
        this.op = op;
        this.left = left;
        this.right = right;
        this.literal = literal;
    }
    isLeaf() {
        return this.op === TokenType.LEAF;
    }
    isAtomic() {
        var _a;
        return (this.isLeaf() || (this.op === TokenType.OP_NOT && ((_a = this.left) === null || _a === void 0 ? void 0 : _a.isLeaf())));
    }
    getLiteralValue() {
        return this.literal || '';
    }
    static CreateAnd(left, right) {
        return new ExpNode(TokenType.BINARY_AND, left, right);
    }
    static CreateNot(exp) {
        return new ExpNode(TokenType.OP_NOT, exp);
    }
    static CreateOr(left, right) {
        return new ExpNode(TokenType.BINARY_OR, left, right);
    }
    static CreateLiteral(lit) {
        return new ExpNode(TokenType.LEAF, null, null, lit);
    }
}
const make = (gen) => {
    const data = gen.next().value;
    switch (data.type) {
        case TokenType.LITERAL:
            return ExpNode.CreateLiteral(data.value);
        case TokenType.OP_NOT:
            return ExpNode.CreateNot(make(gen));
        case TokenType.BINARY_AND: {
            const left = make(gen);
            const right = make(gen);
            return ExpNode.CreateAnd(left, right);
        }
        case TokenType.BINARY_OR: {
            const left = make(gen);
            const right = make(gen);
            return ExpNode.CreateOr(left, right);
        }
    }
    return null;
};
const nodeEvaluator = (tree, literalEvaluator) => {
    if (tree) {
        if (tree.isLeaf()) {
            return literalEvaluator(tree.getLiteralValue());
        }
        if (tree.op === TokenType.OP_NOT) {
            return !nodeEvaluator(tree.left, literalEvaluator);
        }
        if (tree.op === TokenType.BINARY_OR) {
            return (nodeEvaluator(tree.left, literalEvaluator) ||
                nodeEvaluator(tree.right, literalEvaluator));
        }
        if (tree.op === TokenType.BINARY_AND) {
            return (nodeEvaluator(tree.left, literalEvaluator) &&
                nodeEvaluator(tree.right, literalEvaluator));
        }
    }
    return false;
};

const parser = (exp) => {
    const tokens = Tokenizer(exp.replace(/\s/g, ''));
    const polish = PolishNotation(tokens);
    const gen = PolishGenerator(polish);
    const tree = make(gen);
    return tree;
};
const ɵ0$2 = parser;

let configurationAccess = {};
let hasAccessStrategy = (accessName) => of(false);
const ɵ0$3 = hasAccessStrategy;
function setConfigurationAccess(config) {
    configurationAccess = flatten(config, { parse: parser, group: true });
}
function setHasAccessStrategy(accessTest) {
    hasAccessStrategy = accessTest;
}
function parse(expression) {
    const arr = expression.replace(/\s/g, '').split(':');
    return {
        path: arr[0] || '',
        action: arr[1] || '',
    };
}
function canAccessExpression(accessExpression) {
    return of(accessExpression).pipe(map((ae) => ae.replace(/\s/g, '')), switchMap((ae) => nodeEvaluator$1(parser(ae), hasAccessStrategy)));
}
function canAccessPaths(accessPath) {
    const access = Array.isArray(accessPath)
        ? accessPath
        : [accessPath];
    return from(access).pipe(map((ap) => ap.replace(/\s/g, '')), operator(canAccessPath, TokenType.BINARY_OR));
}
function canAccessPath(path) {
    try {
        const access = configurationAccess[path];
        if (!!access) {
            return from(access).pipe(operator((tree) => nodeEvaluator$1(tree, hasAccessStrategy), TokenType.BINARY_OR));
        }
        console.error(`Undefined path ${path}`);
        return of(false);
    }
    catch (e) {
        console.error(e);
        return of(false);
    }
}
function nodeEvaluator$1(tree, literalEvaluator) {
    if (tree) {
        if (tree.isLeaf()) {
            return literalEvaluator(tree.getLiteralValue());
        }
        if (tree.op === TokenType.OP_NOT) {
            return of(tree.left).pipe(operator((tree) => nodeEvaluator$1(tree, literalEvaluator), TokenType.OP_NOT));
        }
        if (tree.op === TokenType.BINARY_OR) {
            return from([tree.left, tree.right]).pipe(operator((tree) => nodeEvaluator$1(tree, literalEvaluator), TokenType.BINARY_OR));
        }
        if (tree.op === TokenType.BINARY_AND) {
            return from([tree.left, tree.right]).pipe(operator((tree) => nodeEvaluator$1(tree, literalEvaluator), TokenType.BINARY_AND));
        }
    }
    return of(false);
}

const ACCESS_CONFIG = new InjectionToken('ACCESS_CONFIG');

class AccessStrategy {
}
class FakeAccessStrategy extends AccessStrategy {
    has(access) {
        return of(false);
    }
}
FakeAccessStrategy.decorators = [
    { type: Injectable }
];

class AccessService {
    constructor(config, accessStrategy) {
        setHasAccessStrategy((accessName) => accessStrategy.has(accessName));
        setConfigurationAccess(config.accesses || {});
    }
    can(accessPaths) {
        return canAccessPaths(accessPaths);
    }
    canExpression(accessExpression) {
        return canAccessExpression(accessExpression);
    }
}
AccessService.ɵprov = ɵɵdefineInjectable({ factory: function AccessService_Factory() { return new AccessService(ɵɵinject(ACCESS_CONFIG), ɵɵinject(AccessStrategy)); }, token: AccessService, providedIn: "root" });
AccessService.decorators = [
    { type: Injectable, args: [{
                providedIn: 'root',
            },] }
];
AccessService.ctorParameters = () => [
    { type: undefined, decorators: [{ type: Inject, args: [ACCESS_CONFIG,] }] },
    { type: AccessStrategy }
];

class AccessDirective {
    constructor(template, viewContainer, accessService, parentAccessDirective) {
        this.template = template;
        this.viewContainer = viewContainer;
        this.accessService = accessService;
        this.parentAccessDirective = parentAccessDirective;
        this.ngxAccess = '';
        this.onDestroy$ = new Subject();
    }
    ngOnInit() {
        if (this.template) {
            let ngxAccess = this.ngxAccess;
            if (this.parentAccessDirective) {
                const { path, action } = parse(this.parentAccessDirective.ngxAccess);
                if (this.ngxAccess) {
                    if (Array.isArray(this.ngxAccess)) {
                        ngxAccess = `${this.ngxAccess.map((access) => access.replace('$', path))}:${action}`;
                    }
                    else {
                        const { path: childPath, action: childAction } = parse(ngxAccess);
                        ngxAccess = `${childPath.replace('$', path)}:${childAction ? childAction : action}`;
                    }
                }
                else {
                    ngxAccess = `${path}:${action}`;
                }
            }
            this.accessService
                .can(ngxAccess)
                .pipe(takeUntil(this.onDestroy$))
                .subscribe((access) => access
                ? this.viewContainer.createEmbeddedView(this.template)
                : this.ngxAccessElse
                    ? this.viewContainer.createEmbeddedView(this.ngxAccessElse)
                    : null);
        }
    }
    ngOnDestroy() {
        this.onDestroy$.next();
        this.onDestroy$.complete();
    }
}
AccessDirective.decorators = [
    { type: Directive, args: [{
                selector: '[ngxAccess]',
            },] }
];
AccessDirective.ctorParameters = () => [
    { type: TemplateRef, decorators: [{ type: Optional }] },
    { type: ViewContainerRef },
    { type: AccessService },
    { type: AccessDirective, decorators: [{ type: Optional }, { type: SkipSelf }, { type: Host }] }
];
AccessDirective.propDecorators = {
    ngxAccess: [{ type: Input }],
    ngxAccessElse: [{ type: Input }]
};

class AccessExpressionDirective {
    constructor(template, viewContainer, accessService) {
        this.template = template;
        this.viewContainer = viewContainer;
        this.accessService = accessService;
        this.ngxAccessExpr = '';
        this.onDestroy$ = new Subject();
    }
    ngOnInit() {
        if (this.template) {
            this.accessService
                .canExpression(this.ngxAccessExpr)
                .pipe(takeUntil(this.onDestroy$))
                .subscribe((access) => access
                ? this.viewContainer.createEmbeddedView(this.template)
                : this.ngxAccessExprElse
                    ? this.viewContainer.createEmbeddedView(this.ngxAccessExprElse)
                    : null);
        }
    }
    ngOnDestroy() {
        this.onDestroy$.next();
        this.onDestroy$.complete();
    }
}
AccessExpressionDirective.decorators = [
    { type: Directive, args: [{
                selector: '[ngxAccessExpr]',
            },] }
];
AccessExpressionDirective.ctorParameters = () => [
    { type: TemplateRef },
    { type: ViewContainerRef },
    { type: AccessService }
];
AccessExpressionDirective.propDecorators = {
    ngxAccessExpr: [{ type: Input }],
    ngxAccessExprElse: [{ type: Input }]
};

class AccessGuard {
    constructor(router, accessService, config) {
        this.router = router;
        this.accessService = accessService;
        this.config = config;
    }
    canActivate(next) {
        return this.accessService
            .can(next.data.accesses)
            .pipe(tap((hasAccess) => !hasAccess &&
            this.router.navigate([next.data.redirect || this.config.redirect])));
    }
}
AccessGuard.ɵprov = ɵɵdefineInjectable({ factory: function AccessGuard_Factory() { return new AccessGuard(ɵɵinject(Router), ɵɵinject(AccessService), ɵɵinject(ACCESS_CONFIG)); }, token: AccessGuard, providedIn: "root" });
AccessGuard.decorators = [
    { type: Injectable, args: [{
                providedIn: 'root',
            },] }
];
AccessGuard.ctorParameters = () => [
    { type: Router },
    { type: AccessService },
    { type: undefined, decorators: [{ type: Inject, args: [ACCESS_CONFIG,] }] }
];

class AccessModule {
    static forRoot(config) {
        return {
            ngModule: AccessModule,
            providers: [
                {
                    provide: ACCESS_CONFIG,
                    useValue: {
                        accesses: config.accesses || {},
                        redirect: config.redirect || '/unauthorized',
                    },
                },
                config.strategy || {
                    provide: AccessStrategy,
                    useClass: FakeAccessStrategy,
                },
            ],
        };
    }
}
AccessModule.decorators = [
    { type: NgModule, args: [{
                declarations: [AccessDirective, AccessExpressionDirective],
                exports: [AccessDirective, AccessExpressionDirective],
            },] }
];

/**
 * Generated bundle index. Do not edit.
 */

export { ACCESS_CONFIG, AccessDirective, AccessExpressionDirective, AccessGuard, AccessModule, AccessService, AccessStrategy, FakeAccessStrategy, AccessDirective as ɵa, AccessService as ɵb, AccessStrategy as ɵc, FakeAccessStrategy as ɵd, AccessExpressionDirective as ɵe };
//# sourceMappingURL=ngx-access.js.map
